@@ -8,11 +8,12 @@ Some features modelled after Racket's builtins for handling procedures.
 Memoize is typical FP (Racket has it in mischief), and flip comes from Haskell.
 """
 
-__all__ = ["memoize", "curry",
+__all__ = ["memoize", "curry", "iscurried",
            "flip", "rotate",
            "apply", "identity", "const", "notf", "andf", "orf",
-           "composer1", "composel1", "composer", "composel",      # *args
-           "composeri1", "composeli1", "composeri", "composeli",  # iterable
+           "composer1", "composel1", "composer1i", "composel1i",  # single arg
+           "composer",  "composel",  "composeri",  "composeli",   # multi-arg
+           "composerc", "composelc", "composerci", "composelci",  # multi-arg w/ curry
            "to1st", "to2nd", "tokth", "tolast", "to"]
 
 from functools import wraps, partial
@@ -71,6 +72,11 @@ def curry(f, *args, **kwargs):
     Also more kwargs can be passed at each step, but they do not affect the
     decision when the function is called.
 
+    For a callable to be curryable, it must be possible to inpect its signature
+    to determine its minimum and maximum positional arities; builtin functions
+    such as ``operator.add`` won't work. In such cases ``UnknownArity`` will
+    be raised.
+
     **Examples**::
 
         @curry
@@ -104,9 +110,9 @@ def curry(f, *args, **kwargs):
     on the remaining positional args::
 
         map_one = lambda f: (curry(foldr))(composer(cons, to1st(f)), nil)
-        assert curry(map_one)(double, (1, 2, 3)) == (2, 4, 6)
+        assert curry(map_one)(double, ll(1, 2, 3)) == ll(2, 4, 6)
 
-    In the above example, ``map_one`` has arity 1, so the tuple ``(1, 2, 3)``
+    In the above example, ``map_one`` has arity 1, so the arg ``ll(1, 2, 3)``
     is extra. The result of ``map_one`` is a callable, so it is then
     invoked on this tuple.
 
@@ -125,6 +131,10 @@ def curry(f, *args, **kwargs):
 
         assert curry(double, 2, "foo") == (4, "foo")
 
+        mymap = lambda f: curry(foldr, composerc(cons, f), nil)
+        add = lambda x, y: x + y
+        assert curry(mymap, add, ll(1, 2, 3), ll(4, 5, 6)) == ll(5, 7, 9)
+
         from functools import partial
         from unpythonic import curry, composel, drop, take
 
@@ -132,6 +142,11 @@ def curry(f, *args, **kwargs):
         look = lambda n1, n2: composel(*with_n((n1, drop), (n2, take)))
         assert tuple(curry(look, 5, 10, range(20))) == tuple(range(5, 15))
     """
+    # trivial case first: prevent stacking curried wrappers
+    if iscurried(f):
+        if args or kwargs:
+            return f(*args, **kwargs)
+        return f
     # TODO: improve: all required name-only args should be present before calling f.
     # Difficult, partial() doesn't remove an already-set kwarg from the signature.
     min_arity, max_arity = arities(f)
@@ -146,7 +161,7 @@ def curry(f, *args, **kwargs):
             if callable(now_result):
                 # curry it now, to sustain the chain in case we still have
                 # too many args.
-                if not hasattr(now_result, "_is_curried_function"):
+                if not iscurried(now_result):
                     now_result = curry(now_result)
                 return now_result(*later_args)
             elif isinstance(now_result, (tuple, list)):
@@ -154,12 +169,16 @@ def curry(f, *args, **kwargs):
             else:
                 return (now_result,) + later_args
         return f(*args, **kwargs)
-    curried._is_curried_function = True  # stash for easy detection
+    curried._is_curried_function = True  # stash for detection
     # curry itself is curried: if we get args, they're the first step
     if args or kwargs:
         return curried(*args, **kwargs)
     return curried
 
+def iscurried(f):
+    """Return whether f is a curried function."""
+    return hasattr(f, "_is_curried_function")
+
 #def curry_simple(f):  # essential idea, without the extra features
 #    min_arity, _ = arities(f)
 #    @wraps(f)
@@ -326,9 +345,47 @@ def _make_compose1(direction):  # "left", "right"
         #   input: a b c
         #   elt = b -> g, acc = a(x) -> f --> a(b(x))
         #   elt = c -> g, acc = a(b(x)) -> f --> a(b(c(x)))
+        # Using reducel is particularly nice here:
+        #  - if fs is empty, we output None
+        #  - if fs contains only one item, we output it as-is
         return unpythonic.it.reducel(compose1_two, fs)  # op(elt, acc)
     return compose1
 
+_compose1_left = _make_compose1("left")
+_compose1_right = _make_compose1("right")
+
+def composer1(*fs):
+    """Like composer, but limited to one-argument functions. Faster.
+
+    Example::
+
+        double = lambda x: 2*x
+        inc    = lambda x: x+1
+        inc_then_double = composer1(double, inc)
+        assert inc_then_double(3) == 8
+    """
+    return composer1i(fs)
+
+def composel1(*fs):
+    """Like composel, but limited to one-argument functions. Faster.
+
+    Example::
+
+        double = lambda x: 2*x
+        inc    = lambda x: x+1
+        double_then_inc = composel1(double, inc)
+        assert double_then_inc(3) == 7
+    """
+    return composel1i(fs)
+
+def composer1i(iterable):  # this is just to insert a docstring
+    """Like composer1, but read the functions from an iterable."""
+    return _compose1_right(iterable)
+
+def composel1i(iterable):
+    """Like composel1, but read the functions from an iterable."""
+    return _compose1_left(iterable)
+
 def _make_compose(direction):  # "left", "right"
     def compose_two(f, g):
         def composed(*args):
@@ -346,43 +403,9 @@ def _make_compose(direction):  # "left", "right"
         return unpythonic.it.reducel(compose_two, fs)  # op(elt, acc)
     return compose
 
-_compose1_left = _make_compose1("left")
-_compose1_right = _make_compose1("right")
 _compose_left = _make_compose("left")
 _compose_right = _make_compose("right")
 
-def composer1(*fs):
-    """Like composer, but limited to one-argument functions. Faster.
-
-    Example::
-
-        double = lambda x: 2*x
-        inc    = lambda x: x+1
-        inc_then_double = composer1(double, inc)
-        assert inc_then_double(3) == 8
-    """
-    return composeri1(fs)
-
-def composeri1(iterable):  # this is just to insert a docstring
-    """Like composeri, but limited to one-argument functions. Faster."""
-    return _compose1_right(iterable)
-
-def composel1(*fs):
-    """Like composel, but limited to one-argument functions. Faster.
-
-    Example::
-
-        double = lambda x: 2*x
-        inc    = lambda x: x+1
-        double_then_inc = composel(double, inc)
-        assert double_then_inc(3) == 7
-    """
-    return composeli1(fs)
-
-def composeli1(iterable):
-    """Like composeli, but limited to one-argument functions. Faster."""
-    return _compose1_left(iterable)
-
 def composer(*fs):
     """Compose functions accepting only positional args. Right to left.
 
@@ -397,10 +420,6 @@ def composer(*fs):
     """
     return composeri(fs)
 
-def composeri(iterable):
-    """Like composer, but takes an iterable of functions to compose."""
-    return _compose_right(iterable)
-
 def composel(*fs):
     """Like composer, but from left to right.
 
@@ -409,10 +428,39 @@ def composel(*fs):
     """
     return composeli(fs)
 
+def composeri(iterable):
+    """Like composer, but read the functions from an iterable."""
+    return _compose_right(iterable)
+
 def composeli(iterable):
-    """Like composel, but takes an iterable of functions to compose."""
+    """Like composel, but read the functions from an iterable."""
     return _compose_left(iterable)
 
+def composerc(*fs):
+    """Like composer, but curry each function before composing.
+
+    With the passthrough in ``curry``, this allows very compact code::
+
+        mymap = lambda f: curry(foldr, composerc(cons, f), nil)
+        assert curry(mymap, double, ll(1, 2, 3)) == ll(2, 4, 6)
+
+        add = lambda x, y: x + y
+        assert curry(mymap, add, ll(1, 2, 3), ll(4, 5, 6)) == ll(5, 7, 9)
+    """
+    return composerci(fs)
+
+def composelc(*fs):
+    """Like composel, but curry each function before composing."""
+    return composelci(fs)
+
+def composerci(iterable):
+    """Like composerc, but read the functions from an iterable."""
+    return composeri(map(curry, iterable))
+
+def composelci(iterable):
+    """Like composelc, but read the functions from an iterable."""
+    return composeli(map(curry, iterable))
+
 # Helpers to insert one-in-one-out functions into multi-arg compose chains
 def tokth(k, f):
     """Return a function to apply f to args[k], pass the rest through.
@@ -441,13 +489,10 @@ def to1st(f):
 
     Example::
 
-        nil = ()
-        def cons(x, l):  # elt, acc
-            return (x,) + l
         def mymap_one(f, sequence):
             f_then_cons = composer(cons, to1st(f))  # args: elt, acc
             return foldr(f_then_cons, nil, sequence)
-        double = lambda x: 2*x
+        double = lambda x: 2 * x
         assert mymap_one(double, (1, 2, 3)) == (2, 4, 6)
     """
     return tokth(0, f)  # this is just a partial() but we want to provide a docstring.
@@ -554,6 +599,10 @@ def test():
     t()
     assert evaluations == 1  # t has no args, so it should have been invoked
 
+    add = lambda x, y: x + y
+    a = curry(add)
+    assert curry(a) is a  # curry wrappers should not stack
+
     # flip
     def f(a, b):
         return (a, b)
