@@ -3,12 +3,15 @@
 """Sequencing constructs - for multi-expression lambdas."""
 
 __all__ = ["begin", "begin0", "lazy_begin", "lazy_begin0",
-           "pipe", "piped", "lazy_piped", "get",
+           "pipe1", "piped1", "lazy_piped1",
+           "pipe", "piped", "getvalue", "lazy_piped", "runpipe",
+           "pipec",  # w/ curry
            "do", "do0", "assign"]
 
 from collections import namedtuple
 from unpythonic.env import env
 from unpythonic.misc import call
+from unpythonic.fun import curry
 
 # evil inspect dependency, used only to provide informative error messages.
 from unpythonic.arity import arity_includes, UnknownArity
@@ -80,7 +83,7 @@ def lazy_begin0(*bodys):
     return out
 
 # sequence one-input, one-output functions
-def pipe(value0, *bodys):
+def pipe1(value0, *bodys):
     """Perform a sequence of operations on an initial value.
 
     Bodys are applied left to right.
@@ -97,15 +100,13 @@ def pipe(value0, *bodys):
 
         x = inc(double(42))  # --> 85
 
-        x = pipe(42, double, inc)  # --> 85
+        x = pipe1(42, double, inc)  # --> 85
 
     but now we don't need to read the source code backwards. This is essentially::
 
-        f = compose(reversed(bodys))
+        f = composel(bodys)
         x = f(42)
 
-    if you have a library that provides ``compose``.
-
     Perhaps the most common alternative in Python is this imperative code::
 
         x = 42
@@ -139,12 +140,13 @@ def pipe(value0, *bodys):
     return x
 
 @call  # make a singleton
-class get:  # sentinel with a nice repr
+class getvalue:  # sentinel with a nice repr
     """Sentinel; pipe into this to exit a shell-like pipe and return the current value."""
     def __repr__(self):
         return "<sentinel for pipe exit>"
+runpipe = getvalue  # same thing, but semantically better name for lazy pipes
 
-class piped:
+class piped1:
     """Shell-like piping syntax.
 
     Eager; apply each function immediately and store the new value.
@@ -157,30 +159,30 @@ class piped:
 
         Return a ``piped`` object, for chainability.
 
-        As the only exception, if ``f`` is the sentinel ``get``, return the
-        current value (useful for exiting the pipe).
+        As the only exception, if ``f`` is the sentinel ``getvalue``,
+        return the current value (useful for exiting the pipe).
 
         A new ``piped`` object is created at each step of piping; the "update"
         is purely functional, nothing is overwritten.
 
         Examples::
 
-            x = piped(42) | double | inc | get
+            x = piped1(42) | double | inc | getvalue
 
-            y = piped(42) | double
-            assert y | inc | get == 85
-            assert y | get == 84  # y is not modified
+            y = piped1(42) | double
+            assert y | inc | getvalue == 85
+            assert y | getvalue == 84  # y is not modified
         """
-        if f is get:
+        if f is getvalue:
             return self._x
         else:
             cls = self.__class__
             return cls(f(self._x))  # functional update
     def __repr__(self):
-        return "<piped at 0x{:x}; value {}>".format(id(self), self._x)
+        return "<piped1 at 0x{:x}; value {}>".format(id(self), self._x)
 
-class lazy_piped:
-    """Like piped, but apply the functions later, at get time.
+class lazy_piped1:
+    """Like piped, but apply the functions later.
 
     This matters if the initial value is mutable:
 
@@ -189,7 +191,7 @@ class lazy_piped:
           the pipeline.
 
         - ``lazy_piped`` just sets up a computation, and performs it when eventually
-          piped into ``get``. The computation always looks up the latest state
+          piped into ``runpipe``. The computation always looks up the latest state
           of the initial value.
 
     Another way to say this is that ``lazy_piped`` looks up the initial value
@@ -205,7 +207,7 @@ class lazy_piped:
     def __or__(self, f):
         """Pipe the value into f; but just plan to do so, don't perform it yet.
 
-        To run the stored computation, pipe into ``get``.
+        To run the stored computation, pipe into ``runpipe``.
 
         Examples::
 
@@ -213,9 +215,9 @@ class lazy_piped:
             def append_succ(l):
                 l.append(l[-1] + 1)
                 return l  # important, handed to the next function in the pipe
-            p = lazy_piped(lst) | append_succ | append_succ  # plan a computation
+            p = lazy_piped1(lst) | append_succ | append_succ  # plan a computation
             assert lst == [1]        # nothing done yet
-            p | get                  # run the computation
+            p | runpipe              # run the computation
             assert lst == [1, 2, 3]  # now the side effect has updated lst.
 
             # lazy pipe as an unfold
@@ -224,13 +226,13 @@ class lazy_piped:
                 a, b = state
                 fibos.append(a)      # store result by side effect
                 return (b, a + b)    # new state, handed to next function in the pipe
-            p = lazy_piped((1, 1))   # load initial state into a lazy pipe
+            p = lazy_piped1((1, 1))  # load initial state into a lazy pipe
             for _ in range(10):      # set up pipeline
                 p = p | nextfibo
-            p | get  # run it
+            p | runpipe
             print(fibos)
         """
-        if f is get:  # compute now
+        if f is runpipe:  # compute now
             v = self._x
             for g in self._funcs:
                 v = g(v)
@@ -240,7 +242,137 @@ class lazy_piped:
             cls = self.__class__
             return cls(x=self._x, _funcs=self._funcs + (f,))
     def __repr__(self):
-        return "<lazy_piped at 0x{:x}; initial value now {}, functions {}>".format(id(self), self._x, self._funcs)
+        return "<lazy_piped1 at 0x{:x}; initial value now {}, functions {}>".format(id(self), self._x, self._funcs)
+
+def pipe(values0, *bodys):
+    """Like pipe1, but with arbitrary number of inputs/outputs at each step.
+
+    The only restriction is that each function must take as many positional
+    arguments as the previous one returns.
+
+    At each step, if the output from a function is a list or a tuple,
+    it is unpacked to the argument list of the next function. Otherwise,
+    we assume the output is intended to be fed to the next function as-is.
+
+    If you only need a one-in-one-out chain, ``pipe1`` is faster.
+
+    Examples::
+
+        a, b = pipe((2, 3),
+                    lambda x, y: (x + 1, 2 * y),
+                    lambda x, y: (x * 2, y + 1))
+        assert (a, b) == (6, 7)
+
+        a, b, c = pipe((2, 3),
+                       lambda x, y: (x + 1, 2 * y, "foo"),
+                       lambda x, y, s: (x * 2, y + 1, "got {}".format(s)))
+        assert (a, b, c) == (6, 7, "got foo")
+
+        a, b = pipe((2, 3),
+                    lambda x, y: (x + 1, 2 * y, "foo"),
+                    lambda x, y, s: (x * 2, y + 1, "got {}".format(s)),
+                    lambda x, y, s: (x + y, s))
+        assert (a, b) == (13, "got foo")
+    """
+    xs = values0
+    for update in bodys:
+        if isinstance(xs, (list, tuple)):
+            xs = update(*xs)
+        else:
+            xs = update(xs)
+    if isinstance(xs, (list, tuple)):
+        return xs if len(xs) > 1 else xs[0]
+    return xs
+
+def pipec(values0, *bodys):
+    """Like pipe, but curry each function before piping.
+
+    Useful with the passthrough in ``curry``. Each function only needs to
+    declare as many of the (leftmost) arguments as it needs to access or modify::
+
+        a, b = pipec((1, 2),
+                     lambda x: x + 1,  # extra args passed through on the right
+                     lambda x, y: (x * 2, y + 1))
+        assert (a, b) == (4, 3)
+    """
+    return pipe(values0, *map(curry, bodys))
+
+class piped:
+    """Like piped1, but for any number of inputs/outputs at each step."""
+    def __init__(self, *xs):
+        """Set up a pipe and load the initial values xs into it."""
+        self._xs = xs
+    def __or__(self, f):
+        """Pipe the values through the function f.
+
+        Example::
+
+            f = lambda x, y: (2*x, y+1)
+            g = lambda x, y: (x+1, 2*y)
+            x = piped(2, 3) | f | g | getvalue  # --> (5, 8)
+        """
+        xs = self._xs
+        if f is getvalue:
+            return xs if len(xs) > 1 else xs[0]
+        else:
+            cls = self.__class__
+            if isinstance(xs, (list, tuple)):
+                newxs = f(*xs)
+            else:
+                newxs = f(xs)
+            if isinstance(newxs, (list, tuple)):
+                return cls(*newxs)
+            else:
+                return cls(newxs)
+    def __repr__(self):
+        return "<piped at 0x{:x}; values {}>".format(id(self), self._xs)
+
+class lazy_piped:
+    """Like lazy_piped1, but for any number of inputs/outputs at each step.
+
+    Examples::
+
+        p1 = lazy_piped(2, 3)
+        p2 = p1 | (lambda x, y: (x + 1, 2 * y, "foo"))
+        p3 = p2 | (lambda x, y, s: (x * 2, y + 1, "got {}".format(s)))
+        p4 = p3 | (lambda x, y, s: (x + y, s))
+        # nothing done yet!
+        assert (p4 | runpipe) == (13, "got foo")
+
+        # lazy pipe as an unfold
+        fibos = []
+        def nextfibo(a, b):    # now two arguments
+            fibos.append(a)
+            return (b, a + b)  # two return values, still expressed as a tuple
+        p = lazy_piped(1, 1)
+        for _ in range(10):
+            p = p | nextfibo
+        p | runpipe
+        print(fibos)
+    """
+    def __init__(self, *xs, _funcs=None):
+        """Set up a lazy pipe and load the initial values xs into it.
+
+        The ``_funcs`` parameter is for internal use.
+        """
+        self._xs = xs
+        self._funcs = _funcs or ()
+    def __or__(self, f):
+        """Pipe the values into f; but just plan to do so, don't perform it yet."""
+        if f is runpipe:  # compute now
+            vs = self._xs
+            for g in self._funcs:
+                if isinstance(vs, (list, tuple)):
+                    vs = g(*vs)
+                else:
+                    vs = g(vs)
+            return vs if len(vs) > 1 else vs[0]
+        else:
+            # just pass on the references to the original xs.
+            cls = self.__class__
+            return cls(*self._xs, _funcs=self._funcs + (f,))
+    def __repr__(self):
+        return "<lazy_piped at 0x{:x}; initial values now {}, functions {}>".format(id(self), self._xs, self._funcs)
 
 # do(): improved begin() that can name intermediate results and refer to them
 DoAssign = namedtuple("DoAssign", "name value")
@@ -355,43 +487,70 @@ def do0(*items):
 
 def test():
     # sequence side effects in a lambda
-    f = lambda x: begin(print("hi"),
-                        42*x)
-    assert f(1) == 42
+    #
+    f1 = lambda x: begin(print("cheeky side effect"), 42*x)
+    assert f1(2) == 84
 
-    g = lambda x: begin0(23*x,
-                         print("hi"))
-    assert g(1) == 23
+    f2 = lambda x: begin0(42*x, print("cheeky side effect"))
+    assert f2(2) == 84
 
-    test_lazy_begin = lambda: lazy_begin(lambda: print("hi"),
-                                         lambda: "the return value")
-    assert test_lazy_begin() == "the return value"
+    f3 = lambda x: lazy_begin(lambda: print("cheeky side effect"),
+                              lambda: 42*x)
+    assert f3(2) == 84
 
-    test_lazy_begin0 = lambda: lazy_begin0(lambda: "the return value",
-                                           lambda: print("hi"))
-    assert test_lazy_begin0() == "the return value"
+    f4 = lambda x: lazy_begin0(lambda: 42*x,
+                               lambda: print("cheeky side effect"))
+    assert f4(2) == 84
 
-    # pipe one-input one-output functions
+    # pipe: sequence functions
+    #
     double = lambda x: 2 * x
     inc    = lambda x: x + 1
-    assert pipe(42, double, inc) == 85
+    assert pipe1(42, double, inc) == 85  # 1-in-1-out
+    assert pipe1(42, inc, double) == 86
+    assert pipe(42, double, inc) == 85   # n-in-m-out, supports also 1-in-1-out
     assert pipe(42, inc, double) == 86
 
-    # with optional shell-like syntax
-    assert piped(42) | double | inc | get == 85
+    # 2-in-2-out
+    a, b = pipe((2, 3),
+                lambda x, y: (x + 1, 2 * y),
+                lambda x, y: (x * 2, y + 1))
+    assert (a, b) == (6, 7)
 
-    y = piped(42) | double
-    assert y | inc | get == 85
-    assert y | get == 84  # y is never modified by the pipe system
+    # 2-in-eventually-3-out
+    a, b, c = pipe((2, 3),
+                   lambda x, y: (x + 1, 2 * y, "foo"),
+                   lambda x, y, z: (x * 2, y + 1, "got {}".format(z)))
+    assert (a, b, c) == (6, 7, "got foo")
 
-    # lazy pipe: compute at get time
+    # 2-in-3-in-between-2-out
+    a, b = pipe((2, 3),
+                lambda x, y: (x + 1, 2 * y, "foo"),
+                lambda x, y, s: (x * 2, y + 1, "got {}".format(s)),
+                lambda x, y, s: (x + y, s))
+    assert (a, b) == (13, "got foo")
+
+    # pipec: curry the functions before running the pipeline
+    a, b = pipec((1, 2),
+                 lambda x: x + 1,  # extra args passed through on the right
+                 lambda x, y: (x * 2, y + 1))
+    assert (a, b) == (4, 3)
+
+    # optional shell-like syntax
+    assert piped1(42) | double | inc | getvalue == 85
+
+    y = piped1(42) | double
+    assert y | inc | getvalue == 85
+    assert y | getvalue == 84  # y is never modified by the pipe system
+
+    # lazy pipe: compute later
     lst = [1]
     def append_succ(l):
         l.append(l[-1] + 1)
         return l  # important, handed to the next function in the pipe
-    p = lazy_piped(lst) | append_succ | append_succ  # plan a computation
+    p = lazy_piped1(lst) | append_succ | append_succ  # plan a computation
     assert lst == [1]        # nothing done yet
-    p | get                  # run the computation
+    p | runpipe              # run the computation
     assert lst == [1, 2, 3]  # now the side effect has updated lst.
 
     # lazy pipe as an unfold
@@ -400,11 +559,42 @@ def test():
         a, b = state
         fibos.append(a)      # store result by side effect
         return (b, a + b)    # new state, handed to next function in the pipe
-    p = lazy_piped((1, 1))   # load initial state into a lazy pipe
+    p = lazy_piped1((1, 1))  # load initial state into a lazy pipe
     for _ in range(10):      # set up pipeline
         p = p | nextfibo
-    p | get  # run it
-    print(fibos)
+    p | runpipe
+    assert fibos == [1, 1, 2, 3, 5, 8, 13, 21, 34, 55]
+
+    # multi-arg version
+    f = lambda x, y: (2*x, y+1)
+    g = lambda x, y: (x+1, 2*y)
+    x = piped(2, 3) | f | g | getvalue  # --> (5, 8)
+    assert x == (5, 8)
+
+    # abuse multi-arg version for single-arg case
+    assert piped(42) | double | inc | getvalue == 85
+
+    # multi-arg lazy pipe
+    p1 = lazy_piped(2, 3)
+    p2 = p1 | (lambda x, y: (x + 1, 2 * y, "foo"))
+    p3 = p2 | (lambda x, y, s: (x * 2, y + 1, "got {}".format(s)))
+    p4 = p3 | (lambda x, y, s: (x + y, s))
+    # nothing done yet, and all computations purely functional:
+    assert (p1 | runpipe) == (2, 3)
+    assert (p2 | runpipe) == (3, 6, "foo")      # runs the chain up to p2
+    assert (p3 | runpipe) == (6, 7, "got foo")  # runs the chain up to p3
+    assert (p4 | runpipe) == (13, "got foo")
+
+    # multi-arg lazy pipe as an unfold
+    fibos = []
+    def nextfibo(a, b):    # now two arguments
+        fibos.append(a)
+        return (b, a + b)  # two return values, still expressed as a tuple
+    p = lazy_piped(1, 1)
+    for _ in range(10):
+        p = p | nextfibo
+    p | runpipe
+    assert fibos == [1, 1, 2, 3, 5, 8, 13, 21, 34, 55]
 
     # do: improved begin() that can name intermediate results
     y = do(assign(x=17),
